#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/*
    2008 - Copyright by Bee <http://www.honeybeenet.altervista.org>.
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#define BEESU_VERSION "2.3"

struct s_argument
{
    char **items;
    unsigned int count;
    unsigned long int size;
};

void add_argument(struct s_argument *arguments, char *argument);
void arg_error(int id, char *me, char *arg);

#define _B(x) ((x)==1)
#define _BF 0
#define _BT 1
typedef int _bool;

const static char const_shell[]="/bin/bash";
const static char const_su[]="/bin/su";
#define READ_BUFFER_SIZE (1024*8)
const static char const_configuration[]="/etc/beesu.conf";

#define NO_EXTERNAL_ARGS (-100)

int main(int argc, char *argv[])
{
    int i;

    char *arg;
    int local_params = 1;
    int external_params = NO_EXTERNAL_ARGS;
    _bool local_verbose = _BF;
    _bool print_configuration_file_and_quit = _BF;

    _bool simulate_login = _BF;
    _bool compress_params = _BT;
    _bool preserve_environment = _BF;
    _bool fast = _BF;
    char *shell = NULL;

    struct s_argument arguments;
    struct s_argument arguments_su;

    char *arguments_su_memory = NULL;

    int exec_result;

    FILE *f_config;
    char *f_read_buffer;
    char f_block_name[10];
    _bool command_in_whitelist = _BF;



    arguments.items = NULL;
    arguments.count = 0;
    arguments.size = 0;
    arguments_su.items = NULL;
    arguments_su.count = 0;
    arguments_su.size = 0;

    for(i = 1; i < argc; i++)
    {
        arg = argv[i];
        if(local_params==1)
        {
            if(strcmp(arg,"-P")==0)
            {
                compress_params = _BF;
            }
            else
            if(strcmp(arg,"-v")==0)
            {
                local_verbose = _BT;
            }
            else
            if(strcmp(arg,"--version")==0)
            {
                printf("beesu " BEESU_VERSION "\n");
                return EXIT_SUCCESS;
            }
            else
            if(strcmp(arg,"--print-conf")==0)
            {
                print_configuration_file_and_quit = _BT;
            }
            else
            if(strcmp(arg,"-c")==0)
            {
                local_params = 100;
                if(argv[i+1] == NULL)
                {
                    arg_error(1, argv[0], arg);
                    return EXIT_FAILURE;
                }
            }
            else
            if(strcmp(arg,"-")==0 || strcmp(arg,"-l")==0 || strcmp(arg,"--login")==0)
            {
                simulate_login = _BT;
            }
            else
            if(strcmp(arg,"-m")==0 || strcmp(arg,"--preserve-environment")==0)
            {
                preserve_environment = _BT;
            }
            else
            if(strcmp(arg,"-f")==0 || strcmp(arg,"--fast")==0)
            {
                fast = _BT;
            }
            else
            if(strcmp(arg,"-s")==0)
            {
                shell = argv[++i];
                if(shell == NULL)
                {
                    shell = (char *)const_shell;
                    i--;
                }
            }
            else
            if(strlen(arg) > 0)
            {
                if(arg[0]=='-')
                {
                    arg_error(0, argv[0], arg);
                    return EXIT_FAILURE;
                }
                else
                {
                    local_params = 0;
                }
            }
        }
        if(local_params==0)
        {
            external_params = i;
            break;//Quit local parsing
        }
        if(local_params==100)
        {
            local_params = 0;
        }
    }

    if(external_params!=NO_EXTERNAL_ARGS || shell!=NULL)
    {
        add_argument(&arguments, "su");
        if(_B(simulate_login))
        {
            add_argument(&arguments, "-l");
        }
        if(_B(preserve_environment))
        {
            add_argument(&arguments, "-m");
        }
        if(_B(fast))
        {
            add_argument(&arguments, "-f");
        }
        if(shell!=NULL)
        {
            add_argument(&arguments, "-s");
            add_argument(&arguments, shell);
        }
        if(external_params != NO_EXTERNAL_ARGS)
        {
            add_argument(&arguments, "-c");
            for(i = external_params; i < argc; i++)
            {
                arg = argv[i];
                add_argument(&arguments_su, arg);
            }
        }
    }

    arguments_su_memory = (char*)malloc(arguments_su.size + arguments_su.count + 1);
    arguments_su_memory[0]='\0';
    //strcpy(arguments_su_memory, "");
    for(i=0;i<arguments_su.count;i++)
    {
        strcat(arguments_su_memory, arguments_su.items[i]);
        if(i!=arguments_su.count-1)
        {
            strcat(arguments_su_memory, " ");
        }
    }

    f_config = fopen(const_configuration, "r");
    if(f_config==NULL)
    {
        fprintf(stderr, "Unable to open beesu configuration file \'%s\'\n",const_configuration);
        free(arguments_su_memory);
        return EXIT_FAILURE;
    }

    f_read_buffer = (char*)malloc(READ_BUFFER_SIZE);

    //f_block_name[0]='\0';
    strcpy(f_block_name, "");
    while(!feof(f_config))
    {
        if(fgets(f_read_buffer,READ_BUFFER_SIZE,f_config)!=NULL)
        {
            while(strlen(f_read_buffer) > 0)
            {
                if(f_read_buffer[strlen(f_read_buffer)-1]=='\n' || f_read_buffer[strlen(f_read_buffer)-1]=='\r')
                {
                    f_read_buffer[strlen(f_read_buffer)-1] = '\0';
                }
                else
                {
                    break;
                }
            }
            if(f_read_buffer[0]==';' || f_read_buffer[0]=='#')
            {
                //it's a comment
            }
            else
            {
                if(strcmp(f_read_buffer,"<allow>")==0)
                {
                    strcpy(f_block_name, "allow");
                    if(_B(print_configuration_file_and_quit))
                        printf("List of allowed commands:\n");
                }
                else
                if(strcmp(f_read_buffer,"</allow>")==0 && strcmp(f_block_name, "allow")==0)
                {
                    strcpy(f_block_name, "");
                    if(_B(print_configuration_file_and_quit))
                        printf("End of allowed commands list.\n");
                }
                else
                {
                    if(strcmp(f_block_name, "allow")==0)
                    {
                        if(strlen(f_read_buffer) > 0)
                        {
                            if(strcmp(f_read_buffer,"*")==0)
                            {
                                command_in_whitelist = _BT;
                                if(_B(print_configuration_file_and_quit))
                                    printf("*ANY*\n");
                            }
                            else
                            {
                                if(_B(print_configuration_file_and_quit))
                                    printf("<%s>\n",f_read_buffer);
                            }
                            if(strcmp(f_read_buffer,arguments_su_memory)==0)
                            {
                                command_in_whitelist = _BT;
                            }
                        }
                    }
                    else
                    {
                        //outside any blocks or unknown...
                    }
                }
            }
        }
    }
    free(f_read_buffer);
    fclose(f_config);

    if(_B(print_configuration_file_and_quit))
    {
        free(arguments_su_memory);
        return EXIT_SUCCESS;
    }

    if(arguments.count==0)
    {
        fprintf(stderr, "Warning: no commands. Nothing to do.\n");
    }
    else
    {
        if(arguments_su.count > 0)
        {
            if(_B(compress_params))
            {
                add_argument(&arguments, arguments_su_memory);
            }
            else
            {
                for(i=0;i<arguments_su.count;i++)
                {
                    add_argument(&arguments, arguments_su.items[i]);
                }
            }
        }
        add_argument(&arguments, NULL);
        if(!_B(command_in_whitelist))
        {
            fprintf(stderr, "Unable to run \'%s\'.\nIt isn\'t in the allowed commands list.\n",
                                arguments_su_memory);
            free(arguments_su_memory);
            return EXIT_FAILURE;
        }
        if(_B(local_verbose))
        {
            fprintf(stderr, "Exec: %s\n",const_su);
            for(i=0;i<arguments.count-1;i++)
            {
                fprintf(stderr, "argv[%i]=<%s>\n",i, arguments.items[i]);
            }
        }
        exec_result = execv(const_su, arguments.items);
    }

    free(arguments_su_memory);

    return EXIT_SUCCESS;
}

void add_argument(struct s_argument *arguments, char *argument)
{
    arguments->items = realloc(arguments->items, (arguments->count+1)*sizeof(char*) );
    arguments->items[arguments->count] = argument;
    arguments->count++;
    if(argument!=NULL)
        arguments->size+=strlen(argument);
    return;
}

void arg_error(int id, char *me, char *arg)
{
    switch(id)
    {
        case 0:
            fprintf(stderr, "%s: invalid option -- \'%s\'\n", me, arg);
            break;
        case 1:
            fprintf(stderr, "%s: option requires an argument -- \'%s\'\n", me, arg);
            break;
    }
    return;
}


